/*
 * grader - event-sched.c
 *
 * Realtime red-black tree event scheduler
 *
 * Copyright (C) 2016  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * Based on Linux v4.7 rbtree implementation by
 * (C) 1999  Andrea Arcangeli <andrea@suse.de>
 * (C) 2002  David Woodhouse <dwmw2@infradead.org>
 * (C) 2012  Michel Lespinasse <walken@google.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <kernel.h>
#include <event-sched.h>
#ifndef CHAR_BIT
#define CHAR_BIT		8
#endif

#if DEBUG
#define debug(fmt...) fprintf(stderr, fmt)
#else
#define debug(fmt...) do { } while (0)
#endif

#if HAVE_STDIO_H && 0
#define info(fmt...) fprintf(stderr, fmt)
#define warn(fmt...) fprintf(stderr, fmt)
#define error(fmt...) fprintf(stderr, fmt)
#define fatal(fmt...) fprintf(stderr, fmt)
#else
#define info(fmt...) do { } while (0)
#define warn(fmt...) do { } while (0)
#define error(fmt...) do { } while (0)
#define fatal(fmt...) do { } while (0)
#endif

#ifndef WARN_ON
#define WARN_ON(cond) \
	do { \
		if ((cond)) { \
			warn("WARNING: in function %s() at %s:%d\n", \
					__FUNCTION__, __FILE__, __LINE__); \
		} \
	} while (0)
#endif

#ifndef BITS_PER_INT
#define BITS_PER_INT (sizeof(int) * CHAR_BIT)
#endif

/*
 * POSITION_BITS - the number of bits of position, up to 24.
 */
#define POSITION_BITS	24

#define POSITION_MASK	((1 << POSITION_BITS) - 1)

/*
 * NR_EVENTS - the maximal number of events in the queue. 4095
 * is the maximal number supported by this implementation. Can
 * be lowered to reduce memory footprint.
 */
#define NR_EVENTS	4095

/*
 * To reduce memory footprint events are identified by 12-bit indices,
 * not pointers.
 */
#define EVENT_MASK	0xfff
#define EVENT_NULL	0

/*
 * The event queue is implemented as red black self-balancing tree.
 * The red black tree has low memory footprint (only 1 additional
 * bit per node - node color) and provides hard realtime guarantees
 * for all operations and fast insertions (compared to AVL trees).
 *
 * This scheduler is designed for LPC176x CPU with 64 KiB of RAM
 * organized as fast 32 KiB RAM region and two slower 16 KiB regions.
 * To distribute queue data to the whole accessible RAM we use
 * three 32-bit arrays (two in 16 KiB regions and one in the 32 KiB
 * region) and one 16-bit (in 32 KiB region).
 *
 *            3                   2                   1                   0
 *          1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
 *         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * rb_a    |   pos[23:16]  |         left          |         right         |
 *         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * rb_b                                    |C|     |         parent        |
 *         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * rb_out0 |                          outputs[31:0]                        |
 *         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * rb_out1 |            pos[15:0]          |        outputs[47:32]         |
 *         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 */

static u32 rb_a[NR_EVENTS + 1];
static u16 rb_b[NR_EVENTS + 1];
static u32 rb_out0[NR_EVENTS + 1] __attribute__((section(".bss.bank0")));
static u32 rb_out1[NR_EVENTS + 1] __attribute__((section(".bss.bank1")));

#define RB_A_RIGHT_SHIFT	0
#define RB_A_LEFT_SHIFT		12
#define RB_A_POS_SHIFT		24
#define RB_A_POS_BITS		8
#define RB_A_POS_MASK		((1 << RB_A_POS_BITS) - 1)

#define RB_B_PARENT_SHIFT	0
#define RB_B_COLOR_SHIFT	15
#define RB_B_COLOR_MASK		1

#define RB_RED		0
#define RB_BLACK	1

#define RB_OUT1_POS_SHIFT	16
#define RB_OUT1_POS_BITS	16
#define RB_OUT1_POS_MASK	((1 << RB_OUT1_POS_BITS) - 1)
#define RB_OUT1_OUTPUTS_BITS	16
#define RB_OUT1_OUTPUTS_MASK	((1 << RB_OUT1_OUTPUTS_BITS) - 1)

#define RB_EMPTY_NODE(node) \
	(rb_b[(node)] == (node))
#define RB_CLEAR_NODE(node) \
	do { rb_b[(node)] = (node); } while (0)

static inline unsigned int rb_left(unsigned int i)
{
	return (rb_a[i] >> RB_A_LEFT_SHIFT) & EVENT_MASK;
}

static inline unsigned int rb_right(unsigned int i)
{
	return (rb_a[i] >> RB_A_RIGHT_SHIFT) & EVENT_MASK;
}

static inline unsigned int rb_parent(unsigned int i)
{
	return (rb_b[i] >> RB_B_PARENT_SHIFT) & EVENT_MASK;
}

static inline unsigned int rb_color(unsigned int i)
{
	return (rb_b[i] >> RB_B_COLOR_SHIFT) & RB_B_COLOR_MASK;
}

#define rb_is_black(i) rb_color(i)
#define rb_is_red(i) (!rb_color(i))

static inline void rb_set_left(unsigned int i, unsigned int p)
{
	unsigned int x = rb_a[i];

	x &= ~(EVENT_MASK << RB_A_LEFT_SHIFT);
	x |= p << RB_A_LEFT_SHIFT;
	rb_a[i] = x;
}

static inline void rb_set_right(unsigned int i, unsigned int p)
{
	unsigned int x = rb_a[i];

	x &= ~(EVENT_MASK << RB_A_RIGHT_SHIFT);
	x |= p << RB_A_RIGHT_SHIFT;
	rb_a[i] = x;
}

static inline void rb_set_parent(unsigned int i, unsigned int p)
{
	unsigned int x = rb_b[i];

	x &= ~(EVENT_MASK << RB_B_PARENT_SHIFT);
	x |= p << RB_B_PARENT_SHIFT;
	rb_b[i] = x;
	WARN_ON(rb_parent(i) != p);
}

static inline void rb_set_parent_color(unsigned int i, unsigned int p,
		int color)
{
	unsigned int x = rb_b[i];

	x &= ~(EVENT_MASK << RB_B_PARENT_SHIFT);
	x &= ~(1 << RB_B_COLOR_SHIFT);
	x |= p << RB_B_PARENT_SHIFT;
	x |= color << RB_B_COLOR_SHIFT;
	rb_b[i] = x;
	WARN_ON(rb_parent(i) != p);
}

static inline void rb_set_black(unsigned int rb)
{
	rb_b[rb] |= RB_BLACK << RB_B_COLOR_SHIFT;
}

static inline unsigned int rb_pos(unsigned int i)
{
	unsigned int pos;

	pos = (rb_a[i] >> RB_A_POS_SHIFT) & RB_A_POS_MASK;
	pos <<= RB_OUT1_POS_BITS;
	pos += (rb_out1[i] >> RB_OUT1_POS_SHIFT) & RB_OUT1_POS_MASK;
	return pos;
}

static inline void rb_set_pos(unsigned int i, unsigned int pos)
{
	unsigned int a = (pos >> RB_OUT1_POS_BITS) & RB_A_POS_MASK;
	unsigned int b = pos & RB_OUT1_POS_MASK;
	unsigned int x;

	x = rb_a[i];
	x &= ~(RB_A_POS_MASK << RB_A_POS_SHIFT);
	x |= a << RB_A_POS_SHIFT;
	rb_a[i] = x;
	
	x = rb_out1[i];
	x &= ~(RB_OUT1_POS_MASK << RB_OUT1_POS_SHIFT);
	x |= b << RB_OUT1_POS_SHIFT;
	rb_out1[i] = x;
}

static inline unsigned long long rb_outputs(unsigned int i)
{
	unsigned long long outputs;

	outputs = rb_out0[i];
	outputs |= rb_out1[i] & RB_OUT1_OUTPUTS_MASK;
	return outputs;
}

#define POS_SHIFT_BITS (BITS_PER_INT - POSITION_BITS)
#define pos_shift(x)	((unsigned int)(x) << POS_SHIFT_BITS)

#define pos_after(a, b) \
	((int) pos_shift(b) - (int) pos_shift(a) < 0)
#define pos_before(a, b) \
	((int) pos_shift(a) - (int) pos_shift(b) < 0)

static void rb_check_node(unsigned int i)
{
	unsigned int left = rb_left(i);
	unsigned int right = rb_right(i);

	if (left)
		WARN_ON(!pos_after(rb_pos(i), rb_pos(left)));
	if (right)
		WARN_ON(pos_before(rb_pos(right), rb_pos(i)));
}

static void __rb_check_tree(unsigned int i)
{
	unsigned int left = rb_left(i);
	unsigned int right = rb_right(i);

	if (!i)
		return;
		
	WARN_ON(rb_parent(i) == i);
	if (left) {
		WARN_ON(left == i);
		WARN_ON(!pos_after(rb_pos(i), rb_pos(left)));
		if (left != i)
			__rb_check_tree(left);
	}
	if (right) {
		WARN_ON(right == i);
		WARN_ON(pos_before(rb_pos(right), rb_pos(i)));
		if (right != i)
			__rb_check_tree(right);
	}
}

static void rb_check_tree(unsigned int i)
{
	__rb_check_tree(i);
}

static unsigned int event_leftmost = EVENT_NULL;
static unsigned int next_pos;

/*
 * Memory management - simple one-way list using rbtree right index
 * and next index. The event_free is an index to the first free
 * entry.
 */
static unsigned int event_free_list = EVENT_NULL;

/**
 * event_alloc - allocate event
 */
static unsigned int event_alloc(void)
{
	unsigned int p = event_free_list;

	if (p < NR_EVENTS) {
		unsigned int next = rb_a[p];

		WARN_ON(next & ~EVENT_MASK);
		event_free_list = rb_a[p] & EVENT_MASK;
		rb_a[p] = 0;
		rb_b[p] = 0;
		rb_out0[p] = 0;
		rb_out1[p] = 0;
	}
	return p;
}

static void event_free(unsigned int p)
{
	rb_a[p] = event_free_list;
	event_free_list = p;
}

static unsigned int rb_level(unsigned int p)
{
	unsigned int level = 0;

	if (p) {
		while ((p = rb_parent(p))) {
			level++;
			if (RB_EMPTY_NODE(p))
				break;
		}
	}

	return level;
}

#define RB_LINK_LEFT	0x1000
static inline unsigned int rb_link_follow(unsigned int link)
{
	unsigned int node = link & EVENT_MASK;

	return (link & RB_LINK_LEFT) ? rb_left(node) : rb_right(node);
}

static inline void rb_link_set(unsigned int link, unsigned int p)
{
	if (link & RB_LINK_LEFT)
		rb_set_left(link & EVENT_MASK, p);
	else
		rb_set_right(link & EVENT_MASK, p);
}

static void event_set_output_bit(unsigned int p, unsigned int output)
{
	if (output < 32)
		rb_out0[p] |= 1 << output;
	else if (output < 52)
		rb_out1[p] |= 1 << (output - 32);
	else
		WARN_ON(1);
}

static void rb_link_node(unsigned int node, unsigned int parent,
		unsigned int link)
{
	debug("linking %d (%d) to %d (%d) as %s\n",
			node, rb_pos(node),
			parent, rb_pos(parent),
			(link & RB_LINK_LEFT) ? "left" : "right");
	rb_set_parent_color(node, parent, RB_RED);
	rb_set_left(node, 0);
	rb_set_right(node, 0);
	rb_link_set(link, node);
}

static inline void __rb_change_child(unsigned int old, unsigned int new,
		unsigned int parent, unsigned int root)
{
	if (parent) {
		if (rb_left(parent) == old)
			rb_set_left(parent, new);
		else
			rb_set_right(parent, new);
	} else {
		/* root uses left for pointer */
		rb_set_left(root, new);
	}
}

/*
 * Helper function for rotations:
 * - old's parent and color get assigned to new
 * - old gets assigned new as a parent and 'color' as a color.
 */
static inline void
__rb_rotate_set_parents(unsigned int old, unsigned int new,
			unsigned int root, int color)
{
	unsigned int parent = rb_parent(old);

	rb_set_parent_color(new, parent, rb_color(old));
	rb_set_parent_color(old, new, color);
	__rb_change_child(old, new, parent, root);
}

static void rb_insert_color(unsigned int node, unsigned int root)
{
	unsigned int parent = rb_parent(node), gparent, tmp;

	while (1) {
		/*
		 * Loop invariant: node is red
		 *
		 * If there is a black parent, we are done.
		 * Otherwise, take some corrective action as we don't
		 * want a red root or two consecutive red nodes.
		 */
		if (!parent) {
			rb_set_parent_color(node, 0, RB_BLACK);
			break;
		} else if (rb_is_black(parent))
			break;

		gparent = rb_parent(parent);

		tmp = rb_right(gparent);
		if (parent != tmp) {	/* parent == gparent->rb_left */
			if (tmp && rb_is_red(tmp)) {
				/*
				 * Case 1 - color flips
				 *
				 *       G            g
				 *      / \          / \
				 *     p   u  -->   P   U
				 *    /            /
				 *   n            n
				 *
				 * However, since g's parent might be red, and
				 * 4) does not allow this, we need to recurse
				 * at g.
				 */
				rb_set_parent_color(tmp, gparent, RB_BLACK);
				rb_set_parent_color(parent, gparent, RB_BLACK);
				node = gparent;
				parent = rb_parent(node);
				rb_set_parent_color(node, parent, RB_RED);
				continue;
			}

			tmp = rb_right(parent);
			if (node == tmp) {
				/*
				 * Case 2 - left rotate at parent
				 *
				 *      G             G
				 *     / \           / \
				 *    p   U  -->    n   U
				 *     \           /
				 *      n         p
				 *
				 * This still leaves us in violation of 4), the
				 * continuation into Case 3 will fix that.
				 */
				tmp = rb_left(node);
				rb_set_right(parent, tmp);
				rb_set_left(node, parent);
				if (tmp)
					rb_set_parent_color(tmp, parent,
							    RB_BLACK);
				rb_set_parent_color(parent, node, RB_RED);
				parent = node;
				tmp = rb_right(node);
			}

			/*
			 * Case 3 - right rotate at gparent
			 *
			 *        G           P
			 *       / \         / \
			 *      p   U  -->  n   g
			 *     /                 \
			 *    n                   U
			 */
			rb_set_left(gparent, tmp); /* == parent->rb_right */
			rb_set_right(parent, gparent);
			if (tmp)
				rb_set_parent_color(tmp, gparent, RB_BLACK);
			__rb_rotate_set_parents(gparent, parent, root, RB_RED);
			break;
		} else {
			tmp = rb_left(gparent);
			if (tmp && rb_is_red(tmp)) {
				/* Case 1 - color flips */
				rb_set_parent_color(tmp, gparent, RB_BLACK);
				rb_set_parent_color(parent, gparent, RB_BLACK);
				node = gparent;
				parent = rb_parent(node);
				rb_set_parent_color(node, parent, RB_RED);
				continue;
			}

			tmp = rb_left(parent);
			if (node == tmp) {
				/* Case 2 - right rotate at parent */
				tmp = rb_right(node);
				rb_set_left(parent, tmp);
				rb_set_right(node, parent);
				if (tmp)
					rb_set_parent_color(tmp, parent,
							    RB_BLACK);
				rb_set_parent_color(parent, node, RB_RED);
				parent = node;
				tmp = rb_left(node);
			}

			/* Case 3 - left rotate at gparent */
			rb_set_right(gparent, tmp); /* == parent->rb_left */
			rb_set_left(parent, gparent);
			if (tmp)
				rb_set_parent_color(tmp, gparent, RB_BLACK);
			__rb_rotate_set_parents(gparent, parent, root, RB_RED);
			break;
		}
	}
}

void cup_queue_event(unsigned int pos, unsigned int output)
{
	unsigned int p;
	unsigned int link = RB_LINK_LEFT;
	unsigned int parent = 0;
	unsigned int node;
	int leftmost = 1;

	while (rb_link_follow(link)) {
		parent = rb_link_follow(link);

		p = rb_pos(parent);
		if (pos_before(pos, p)) {
			link = parent | RB_LINK_LEFT;
		} else if (pos_after(pos, p)) {
			link = parent;
			leftmost = 0;
		} else {
			debug("hit, pos: %d, setting bit %d\n", pos, output);
			event_set_output_bit(parent, output);
			return;
		}
	}

	node = event_alloc();

	if (leftmost) {
		event_leftmost = node;
		next_pos = pos;
	}
	rb_set_pos(node, pos);
	event_set_output_bit(node, output);
	rb_link_node(node, parent, link);
	rb_insert_color(node, 0);

	rb_check_tree(rb_link_follow(RB_LINK_LEFT));
}

static unsigned int rb_next(unsigned int node)
{
	unsigned int parent;

	if (RB_EMPTY_NODE(node))
		return 0;

	rb_check_node(node);

	/*
	 * If we have a right-hand child, go down and then left as far
	 * as we can.
	 */
	if (rb_right(node)) {
		node = rb_right(node);
		rb_check_node(node);
		while (rb_left(node)) {
			node = rb_left(node);
			rb_check_node(node);
		}
		return node;
	}

	/*
	 * No right-hand children. Everything down and left is smaller than us,
	 * so any 'next' node must be in the general direction of our parent.
	 * Go up the tree; any time the ancestor is a right-hand child of its
	 * parent, keep going up. First time it's a left-hand child of its
	 * parent, said parent is our 'next' node.
	 */
	while ((parent = rb_parent(node)) && node == rb_right(parent))
		node = parent;

	return parent;
}

/*
 * Inline version for rb_erase() use - we want to be able to inline
 * and eliminate the dummy_rotate callback there
 */
static void __rb_erase_color(unsigned int parent, unsigned int root)
{
	unsigned int node = 0, sibling, tmp1, tmp2;

	while (1) {
		/*
		 * Loop invariants:
		 * - node is black (or NULL on first iteration)
		 * - node is not the root (parent is not NULL)
		 * - All leaf paths going through parent and node have a
		 *   black node count that is 1 lower than other leaf paths.
		 */
		sibling = rb_right(parent);
		if (node != sibling) {	/* node == parent->rb_left */
			if (rb_is_red(sibling)) {
				/*
				 * Case 1 - left rotate at parent
				 *
				 *     P               S
				 *    / \             / \
				 *   N   s    -->    p   Sr
				 *      / \         / \
				 *     Sl  Sr      N   Sl
				 */
				tmp1 = rb_left(sibling);
				rb_set_right(parent, tmp1);
				rb_set_left(sibling, parent);
				rb_set_parent_color(tmp1, parent, RB_BLACK);
				__rb_rotate_set_parents(parent, sibling, root,
							RB_RED);
				sibling = tmp1;
			}
			tmp1 = rb_right(sibling);
			if (!tmp1 || rb_is_black(tmp1)) {
				tmp2 = rb_left(sibling);
				if (!tmp2 || rb_is_black(tmp2)) {
					/*
					 * Case 2 - sibling color flip
					 * (p could be either color here)
					 *
					 *    (p)           (p)
					 *    / \           / \
					 *   N   S    -->  N   s
					 *      / \           / \
					 *     Sl  Sr        Sl  Sr
					 *
					 * This leaves us violating 5) which
					 * can be fixed by flipping p to black
					 * if it was red, or by recursing at p.
					 * p is red when coming from Case 1.
					 */
					rb_set_parent_color(sibling, parent,
							    RB_RED);
					if (rb_is_red(parent))
						rb_set_black(parent);
					else {
						node = parent;
						parent = rb_parent(node);
						if (parent)
							continue;
					}
					break;
				}
				/*
				 * Case 3 - right rotate at sibling
				 * (p could be either color here)
				 *
				 *   (p)           (p)
				 *   / \           / \
				 *  N   S    -->  N   Sl
				 *     / \             \
				 *    sl  Sr            s
				 *                       \
				 *                        Sr
				 */
				tmp1 = rb_right(tmp2);
				rb_set_left(sibling, tmp1);
				rb_set_right(tmp2, sibling);
				rb_set_right(parent, tmp2);
				if (tmp1)
					rb_set_parent_color(tmp1, sibling,
							    RB_BLACK);
				tmp1 = sibling;
				sibling = tmp2;
			}
			/*
			 * Case 4 - left rotate at parent + color flips
			 * (p and sl could be either color here.
			 *  After rotation, p becomes black, s acquires
			 *  p's color, and sl keeps its color)
			 *
			 *      (p)             (s)
			 *      / \             / \
			 *     N   S     -->   P   Sr
			 *        / \         / \
			 *      (sl) sr      N  (sl)
			 */
			tmp2 = rb_left(sibling);
			rb_set_right(parent, tmp2);
			rb_set_left(sibling, parent);
			rb_set_parent_color(tmp1, sibling, RB_BLACK);
			if (tmp2)
				rb_set_parent(tmp2, parent);
			__rb_rotate_set_parents(parent, sibling, root,
						RB_BLACK);
			break;
		} else {
			sibling = rb_left(parent);
			if (rb_is_red(sibling)) {
				/* Case 1 - right rotate at parent */
				tmp1 = rb_right(sibling);
				rb_set_left(parent, tmp1);
				rb_set_right(sibling, parent);
				rb_set_parent_color(tmp1, parent, RB_BLACK);
				__rb_rotate_set_parents(parent, sibling, root,
							RB_RED);
				sibling = tmp1;
				rb_check_tree(rb_link_follow(RB_LINK_LEFT));
			}
			tmp1 = rb_left(sibling);
			if (!tmp1 || rb_is_black(tmp1)) {
				tmp2 = rb_right(sibling);
				if (!tmp2 || rb_is_black(tmp2)) {
					/* Case 2 - sibling color flip */
					rb_set_parent_color(sibling, parent,
							    RB_RED);
					if (rb_is_red(parent)) {
						rb_set_black(parent);
					} else {
						node = parent;
						parent = rb_parent(node);
						if (parent)
							continue;
					}
					break;
				}
				/* Case 3 - right rotate at sibling */
				tmp1 = rb_left(tmp2);
				rb_set_right(sibling, tmp1);
				rb_set_left(tmp2, sibling);
				rb_set_left(parent, tmp2);
				if (tmp1)
					rb_set_parent_color(tmp1, sibling,
							    RB_BLACK);
				tmp1 = sibling;
				sibling = tmp2;
				rb_check_tree(rb_link_follow(RB_LINK_LEFT));
			}
			rb_check_tree(rb_link_follow(RB_LINK_LEFT));
			/* Case 4 - left rotate at parent + color flips */
			tmp2 = rb_right(sibling);
			rb_set_left(parent, tmp2);
			rb_set_right(sibling, parent);
			rb_set_parent_color(tmp1, sibling, RB_BLACK);
			if (tmp2)
				rb_set_parent(tmp2, parent);
			__rb_rotate_set_parents(parent, sibling, root,
						RB_BLACK);
			rb_check_tree(rb_link_follow(RB_LINK_LEFT));
			break;
		}
	}
}

static void rb_erase(unsigned int node, unsigned int root)
{
	unsigned int child = rb_right(node);
	unsigned int tmp = rb_left(node);
	unsigned int parent, rebalance;
	unsigned int pc_parent;
	int pc_color;

	if (!tmp) {
		/*
		 * Case 1: node to erase has no more than 1 child (easy!)
		 *
		 * Note that if there is one child it must be red due to 5)
		 * and node must be black due to 4). We adjust colors locally
		 * so as to bypass __rb_erase_color() later on.
		 */
		pc_parent = rb_parent(node);
		pc_color = rb_color(node);
		parent = pc_parent;
		__rb_change_child(node, child, parent, root);
		if (child) {
			rb_set_parent_color(child, pc_parent, pc_color);
			rebalance = 0;
		} else {
			rebalance = pc_color ? parent : 0;
		}
		tmp = parent;
	} else if (!child) {
		/* Still case 1, but this time the child is node->rb_left */
		pc_parent = rb_parent(node);
		pc_color = rb_color(node);
		rb_set_parent_color(tmp, pc_parent, pc_color);
		parent = pc_parent;
		__rb_change_child(node, tmp, parent, root);
		rebalance = 0;
		tmp = parent;
	} else {
		unsigned int successor = child, child2;

		tmp = rb_left(child);
		if (!tmp) {
			/*
			 * Case 2: node's successor is its right child
			 *
			 *    (n)          (s)
			 *    / \          / \
			 *  (x) (s)  ->  (x) (c)
			 *        \
			 *        (c)
			 */
			parent = successor;
			child2 = rb_right(successor);
		} else {
			/*
			 * Case 3: node's successor is leftmost under
			 * node's right child subtree
			 *
			 *    (n)          (s)
			 *    / \          / \
			 *  (x) (y)  ->  (x) (y)
			 *      /            /
			 *    (p)          (p)
			 *    /            /
			 *  (s)          (c)
			 *    \
			 *    (c)
			 */
			do {
				parent = successor;
				successor = tmp;
				tmp = rb_left(tmp);
			} while (tmp);
			child2 = rb_right(successor);
			rb_set_left(parent, child2);
			rb_set_right(successor, child);
			rb_set_parent(child, successor);
		}

		tmp = rb_left(node);
		rb_set_left(successor, tmp);
		rb_set_parent(tmp, successor);

		pc_parent = rb_parent(node);
		pc_color = rb_color(node);
		tmp = pc_parent;
		__rb_change_child(node, successor, tmp, root);

		if (child2) {
			rb_set_parent_color(successor, pc_parent, pc_color);
			rb_set_parent_color(child2, parent, RB_BLACK);
			rebalance = 0;
		} else {
			int is_black = rb_is_black(successor);
			rb_set_parent_color(successor, pc_parent, pc_color);
			rebalance = is_black ? parent : 0;
		}
		tmp = successor;
	}

	rb_check_tree(rb_link_follow(RB_LINK_LEFT));
	if (rebalance)
		__rb_erase_color(rebalance, root);
	rb_check_tree(rb_link_follow(RB_LINK_LEFT));
}

unsigned long long event_schedule(unsigned int pos)
{
	unsigned int node = event_leftmost;
	unsigned int next;
	unsigned long long ret;

	if (!node)
		return 0;

	if (pos_before(pos, next_pos))
		return 0;

	WARN_ON(next_pos != rb_pos(node));
	ret = rb_outputs(node);
	info("executing event %d pos: %d, current %d, %llx, level: %d\n",
			node, next_pos, pos, ret, rb_level(node));

	next = rb_next(node);
	next_pos = 0;
	if (next)
		next_pos = rb_pos(next);
	event_leftmost = next;

	rb_erase(node, 0);
	event_free(node);
	return ret;
}

void cupsched_init(void)
{
	unsigned int i;

	rb_set_parent(0, 0);
	for (i = 1; i < NR_EVENTS; i++)
		event_free(i);
}

#if 0
int main(void)
{
	int i;

	cupsched_init();
		
	cup_queue_event(1, 0);
	fprintf(stderr, "0\n");
	event_schedule(0);
	fprintf(stderr, "1\n");
	event_schedule(1);
	fprintf(stderr, "2\n");
	event_schedule(2);

	exit(0);

	cup_queue_event((unsigned int) -30 & POSITION_MASK, 0);
	cup_queue_event((unsigned int) -70 & POSITION_MASK, 0);
	cup_queue_event(9, 0);
	cup_queue_event(8, 0);
	cup_queue_event(7, 0);
	cup_queue_event(6, 0);
	cup_queue_event(5, 0);
	cup_queue_event(4, 0);
	cup_queue_event(3, 0);
	cup_queue_event(2, 0);
	cup_queue_event(1, 0);

	for (i = -300; i < 10; i++)
		event_schedule((unsigned int) i & POSITION_MASK);
	
	for (; i < 1000; i++) {
		cup_queue_event(i + 100 * drand48(), 0);
		event_schedule((unsigned int) i & POSITION_MASK);
	}

	return 0;
}
#endif
