/*
 * grader-io - grtp-gen.c
 *
 * GRTP frame generator
 *
 * Copyright (C) 2016  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <grtp.h>
#include <crc14.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <string.h>

static const char *device;
static speed_t speed = B115200;
static unsigned long long data;
static unsigned long frames = 1;

static const char copyright_string[] =
"grtp-gen " PACKAGE_VERSION ".\n"
"Copyright (C) 2016  Krzysztof Mazur <krzysiek@podlesie.net>.\n";

static void show_copyright(void)
{
	puts(copyright_string);
}

static const char usage[] =
"usage: %s [options] [DATA]\n"
"\n"
"Generate GRTP frames\n"
"\n"
"Options:\n"
"  -V, --version              show version\n"
"  -h, --help                 show this help\n"
"  -n, --number=N             generate N frames\n"
"\n"
"Report bugs to krzysiek@podlesie.net/\n";

static void show_usage(char *name)
{
	show_copyright();

	printf(usage, (name == NULL) ? "grtp-gen" : name);
}

static const char license[] =
"This program comes with ABSOLUTELY NO WARRANTY; see COPYING for details.\n"
"This is free software, and you are welcome to redistribute it\n"
"under certain conditions; see COPYING for details.\n";

static void show_version(void)
{
	show_copyright();

	puts(license);
}

static void parse_command_line(int argc, char **args)
{
	static struct option long_options[] = {
		{ "device", 1, NULL, 'd' },
		{ "help", 0, NULL, 'h' },
		{ "number", 1, NULL, 'n' },
		{ "version", 0, NULL, 'V' },
		{ NULL, 0, NULL, 0 },
	};
	int option_index = 0;
	int c;

	while (1) {
		c = getopt_long(argc, args, "d:hn:V",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'd':
			free((void *) device);
			device = strdup(optarg);
			break;
		case 'h':
			show_usage(args[0]);
			exit(EXIT_SUCCESS);
			break;
		case 'n':
			frames = atol(optarg);
			break;
		case 'V':
			show_version();
			exit(EXIT_SUCCESS);
			break;
		default:
			fprintf(stderr, "Unknown option.\n");
			exit(EXIT_FAILURE);
		}
	}

	if (optind >= argc)
		return;
	data = strtoll(args[optind++], NULL, 0);
}

static void grtp_send_data(unsigned long long x, FILE *f)
{
	unsigned char msg[GRTP_FRAME_LEN];
	unsigned int bits = 0;
	unsigned int crc;
	unsigned int i;

	msg[0] = GRTP_SYNC | ((x >> 0) & 0x7f);
	msg[1] = (x >> 8) & 0x7f;
	msg[2] = (x >> 16) & 0x7f;
	msg[3] = (x >> 24) & 0x7f;
	msg[4] = (x >> 32) & 0x7f;
	msg[5] = (x >> 40) & 0x7f;
	msg[6] = (x >> 48) & 0x7f;
	for (i = 0; i < 7; i++) {
		x <<= 8;
		bits <<= 1;
		if (x & 0x8000000000000000ULL)
			bits |= 0x01;
	}
	msg[7] = bits;

	crc = crc14(GRTP_CHECKSUM_INIT, msg, sizeof(msg) - 2);
	crc = ~crc;
	msg[GRTP_FRAME_LEN - 2] = (crc >> 0) & 0x7f;
	msg[GRTP_FRAME_LEN - 1] = (crc >> 7) & 0x7f;
	fwrite(msg, GRTP_FRAME_LEN, 1, f);
}

int main(int argc, char **args)
{
	struct termios mode;
	FILE *f = stdout;

	parse_command_line(argc, args);

	if (device) {
		int fd;

		fd = open(device, O_RDWR | O_NOCTTY);
		if (fd == -1) {
			perror("open");
			exit(EXIT_FAILURE);
		}

		memset(&mode, 0, sizeof(struct termios));
		cfmakeraw(&mode);
		mode.c_cflag |= CLOCAL | CREAD | CS8;
		mode.c_oflag = 0;
		cfsetispeed(&mode, speed);
		cfsetospeed(&mode, speed);
		tcflush(fd, TCIFLUSH);
		tcsetattr(fd, TCSANOW, &mode);

		f = fdopen(fd, "w");
		if (!f) {
			perror("fdopen");
			exit(EXIT_FAILURE);
		}
	}

	if (!frames) {
		int c;

		while ((c = getchar()) != EOF) {
			unsigned long long cdata = c;

			grtp_send_data((cdata << 48) | data, f);
		}
	} else {
		while (frames-- > 0)
			grtp_send_data(data, f);
	}
	return 0;
}
