/*
 * grader - grtp.c
 *
 * Grader Realtime Transport Protocol
 *
 * Copyright (C) 2016  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <kernel.h>
#include <tty.h>
#include <crc14.h>
#include <grtp.h>

#define GRTP_MAX	4

#define GRTP_BUF_SIZE	8

struct grtp {
	struct tty_struct tty;
	struct tty_struct *parent_tty;
	unsigned int len;
	void (*callback)(struct tty_struct *tty, const u32 *frame);
	u32 frame[(GRTP_FRAME_LEN + 3) / 4];
	u16 crc;

	unsigned char rx_buf[GRTP_BUF_SIZE];
	unsigned int rx_head;
	unsigned int rx_tail;
};

static struct grtp grtp_table[GRTP_MAX];

static inline u32 grtp_decode_word(u32 x, unsigned int bits)
{
	static u32 decode_tab[16] __attribute__((section(".data"))) = {
		0x00000000,	/* 0x0 */
		0x00000080,	/* 0x1 */
		0x00008000,	/* 0x2 */
		0x00008080,	/* 0x3 */
		0x00800000,	/* 0x4 */
		0x00800080,	/* 0x5 */
		0x00808000,	/* 0x6 */
		0x00808080,	/* 0x7 */
		0x80000000,	/* 0x8 */
		0x80000080,	/* 0x9 */
		0x80008000,	/* 0xa */
		0x80008080,	/* 0xb */
		0x80800000,	/* 0xc */
		0x80800080,	/* 0xd */
		0x80808000,	/* 0xe */
		0x80808080,	/* 0xf */
	};

	return x + decode_tab[bits];
}

#define le32_to_cpu(x)	(x)
#define cpu_to_le32(x)	(x)

static void grtp_receive_frame(struct grtp *g) __attribute__((hot));

static void grtp_receive_frame(struct grtp *g)
{
	u32 *frame = g->frame;
	unsigned int a;
	unsigned int b;
	unsigned int bits;
	unsigned int head;
	unsigned int next;

	a = le32_to_cpu(frame[0]);
	b = le32_to_cpu(frame[1]);
	a &= ~GRTP_SYNC;
	bits = b >> 24;
	a = grtp_decode_word(a, bits & 0xf);
	b = grtp_decode_word(b, bits >> 4);
	frame[0] = cpu_to_le32(a);
	frame[1] = cpu_to_le32(b);

	head = g->rx_head;
	next = (head + 1) & (GRTP_BUF_SIZE - 1);
	if (next != g->rx_tail) {
		g->rx_buf[head] = ((u8 *) frame)[6];
		g->rx_head = next;
	}

	g->callback(&g->tty, frame);
}

static void grtp_receive_buf(struct tty_struct *tty, const unsigned char *cp,
		char *fp, int count)
	__attribute__((hot));

static void grtp_receive_buf(struct tty_struct *tty, const unsigned char *cp,
		char *fp, int count)
{
	struct grtp *g = tty->disc_data;
	unsigned int len = g->len;

	if (!count)
		return;

	do {
		unsigned char c;
		u8 *frame = (u8 *) g->frame;

		/*
		 * if case of receive error drop character and current frame
		 */
		if (fp && *fp++) {
			len = 0;
			cp++;
			continue;
		}

		c = *cp;
		if (c & GRTP_SYNC) {
			len = 0;
			frame[len++] = c;
			g->crc = crc14_byte(GRTP_CHECKSUM_INIT, c);
		} else if (len) {
			frame[len++] = c;
			g->crc = crc14_byte(g->crc, c);
			if (len >= GRTP_FRAME_LEN) {
				if (g->crc == GRTP_CHECKSUM_CHECK)
					grtp_receive_frame(g);
				len = 0;
			}
		}
		cp++;
		count--;
	} while (count > 0);
	g->len = len;
}

void grtp_poll(struct tty_struct *tty)
{
	struct grtp *g = container_of(tty, struct grtp, tty);
	unsigned int tail;
	unsigned int head;

	if (!g->tty.ldisc_ops)
		return;

	tail = g->rx_tail;
	head = ACCESS_ONCE(g->rx_head);
	while (tail != head) {
		tty_receive_buf(&g->tty, g->rx_buf + tail, NULL, 1);
		tail = (tail + 1) & (GRTP_BUF_SIZE - 1);
	}
	ACCESS_ONCE(g->rx_tail) = tail;
}

int grtp_xmit(struct tty_struct *tty, const unsigned char *frame)
{
	struct grtp *g = container_of(tty, struct grtp, tty);
	unsigned char msg[GRTP_FRAME_LEN];
	unsigned int bits = 0;
	unsigned int crc;
	unsigned int i;

	i = 6;
	do {
		bits <<= 1;
		msg[i] = frame[i] & 0x7f;
		if (frame[i] & 0x80)
			bits |= 0x01;
	} while (i-- > 0);
	msg[0] |= GRTP_SYNC;
	msg[7] = bits;

	crc = crc14(GRTP_CHECKSUM_INIT, msg, sizeof(msg) - 2);
	crc = ~crc;
	msg[GRTP_FRAME_LEN - 2] = (crc >> 0) & 0x7f;
	msg[GRTP_FRAME_LEN - 1] = (crc >> 7) & 0x7f;

	g->parent_tty->ops->write(g->parent_tty, msg, sizeof(msg));
	return 0;
}

static int grtp_tty_write(struct tty_struct *tty, const unsigned char *buf,
		int count)
{
	return -EINVAL;
}

static const struct tty_operations grtp_tty_operations = {
	.write		= grtp_tty_write,
};

static const struct tty_ldisc_ops grtp_ldisc_ops = {
	.receive_buf		= grtp_receive_buf,
};

struct tty_struct *grtp_init(struct tty_struct *tty,
		void (*callback)(struct tty_struct *tty, const u32 *frame))
{
	struct grtp *g;

	for (g = grtp_table; g < grtp_table + GRTP_MAX; g++) {
		if (!g->tty.ops) {
			g->tty.ops = &grtp_tty_operations;
			g->callback = callback;
			g->parent_tty = tty;

			tty_set_ldisc(tty, &grtp_ldisc_ops, g);
			return &g->tty;
		}
	}
	return NULL;
}
